#include "precomp.h"
#include <initguid.h>
#include <wdmguid.h>

#if defined (EVENT_TRACING)
#ifdef offsetof
#undef offsetof
#endif
#include "pdo.tmh"
#endif 

#define MAX_ID_LEN 80

NTSTATUS
EvtDeviceProcessQueryInterfaceRequest(
	IN WDFDEVICE  Device,
	IN LPGUID  InterfaceType,
	IN OUT PINTERFACE  ExposedInterface,
	IN OUT PVOID  ExposedInterfaceSpecificData
	)
{
	PPDO_DEVICE_DATA p_pdo = PdoGetData(Device);
	PFDO_DEVICE_DATA p_fdo  = p_pdo->p_fdo;
	PMLX4_BUS_IB_INTERFACE p_ib_ifc = (PMLX4_BUS_IB_INTERFACE)ExposedInterface;

	UNUSED_PARAM(InterfaceType);
	UNUSED_PARAM(ExposedInterfaceSpecificData);

	if (p_fdo->pci_dev.dev) {
		p_ib_ifc->p_ibdev = p_fdo->bus_ib_ifc.p_ibdev;
		p_ib_ifc->pmlx4_dev = p_fdo->bus_ib_ifc.pmlx4_dev;
		p_ib_ifc->is_livefish = p_fdo->bus_ib_ifc.is_livefish;
		return STATUS_SUCCESS;
	}
	else {
		p_ib_ifc->p_ibdev = NULL;
		p_ib_ifc->pmlx4_dev = NULL;
		p_ib_ifc->is_livefish = FALSE;
		return STATUS_UNSUCCESSFUL;
	}

}

NTSTATUS
create_pdo(
	__in WDFDEVICE  Device,
	__in PWCHAR     HardwareIds,
	__in PWCHAR     DeviceDescription,
	__in ULONG      SerialNo,
	__in PWCHAR     Location
)
/*++

Routine Description:

    This routine creates and initialize a PDO.

Arguments:

Return Value:

    NT Status code.

--*/
{
	NTSTATUS                    status;
	PWDFDEVICE_INIT             pDeviceInit = NULL;
	PPDO_DEVICE_DATA            p_pdo = NULL;
	PFDO_DEVICE_DATA            p_fdo;
	WDFDEVICE                   hChild = NULL;
	WDF_OBJECT_ATTRIBUTES       pdoAttributes;
	WDF_DEVICE_PNP_CAPABILITIES pnpCaps;
	WDF_DEVICE_POWER_CAPABILITIES powerCaps;
	UNICODE_STRING compatId;
	UNICODE_STRING deviceId;
	DECLARE_UNICODE_STRING_SIZE(buffer, MAX_ID_LEN);
	DECLARE_UNICODE_STRING_SIZE(deviceLocation, MAX_ID_LEN);

	MLX4_PRINT(TRACE_LEVEL_INFORMATION, MLX4_DBG_DRV, ("Entered CreatePdo\n"));

	RtlInitUnicodeString(&compatId, HardwareIds);
	PAGED_CODE();

	//
	// Allocate a WDFDEVICE_INIT structure and set the properties
	// so that we can create a device object for the child.
	//
	pDeviceInit = WdfPdoInitAllocate(Device);

	if (pDeviceInit == NULL) {
		status = STATUS_INSUFFICIENT_RESOURCES;
		goto Cleanup;
	}

	//
	// Set DeviceType
	//
	WdfDeviceInitSetDeviceType(pDeviceInit, FILE_DEVICE_BUS_EXTENDER);

	//
	// Provide DeviceID, HardwareIDs, CompatibleIDs and InstanceId
	//
	RtlInitUnicodeString(&deviceId,HardwareIds);

	status = WdfPdoInitAssignDeviceID(pDeviceInit, &deviceId);
	if (!NT_SUCCESS(status)) {
		goto Cleanup;
	}

	//
	// Note same string  is used to initialize hardware id too
	//
	status = WdfPdoInitAddHardwareID(pDeviceInit, &deviceId);
	if (!NT_SUCCESS(status)) {
		goto Cleanup;
	}

	status = WdfPdoInitAddCompatibleID(pDeviceInit, &compatId);
	if (!NT_SUCCESS(status)) {
		goto Cleanup;
	}

	status =  RtlUnicodeStringPrintf(&buffer, L"%02d", SerialNo);
	if (!NT_SUCCESS(status)) {
		goto Cleanup;
	}

	status = WdfPdoInitAssignInstanceID(pDeviceInit, &buffer);
	if (!NT_SUCCESS(status)) {
		goto Cleanup;
	}

	//
	// Provide a description about the device. This text is usually read from
	// the device. In the case of USB device, this text comes from the string
	// descriptor. This text is displayed momentarily by the PnP manager while
	// it's looking for a matching INF. If it finds one, it uses the Device
	// Description from the INF file or the friendly name created by
	// coinstallers to display in the device manager. FriendlyName takes
	// precedence over the DeviceDesc from the INF file.
	//
	status = RtlUnicodeStringPrintf(&buffer,DeviceDescription  , SerialNo );
	
	if (!NT_SUCCESS(status)) {
		goto Cleanup;
	}

	//
	// You can call WdfPdoInitAddDeviceText multiple times, adding device
	// text for multiple locales. When the system displays the text, it
	// chooses the text that matches the current locale, if available.
	// Otherwise it will use the string for the default locale.
	// The driver can specify the driver's default locale by calling
	// WdfPdoInitSetDefaultLocale.
	//
	status =  RtlUnicodeStringPrintf(&deviceLocation, L"MLX4 %s", Location+4);
	if (!NT_SUCCESS(status)) {
		goto Cleanup;
	}

	status = WdfPdoInitAddDeviceText(pDeviceInit,
		&buffer, &deviceLocation, 0x409);
	if (!NT_SUCCESS(status)) {
		goto Cleanup;
	}

	WdfPdoInitSetDefaultLocale(pDeviceInit, 0x409);

	//
	// Initialize the attributes to specify the size of PDO device extension.
	// All the state information private to the PDO will be tracked here.
	//
	WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&pdoAttributes, PDO_DEVICE_DATA);

	status = WdfDeviceCreate(&pDeviceInit, &pdoAttributes, &hChild);
	if (!NT_SUCCESS(status)) {
		goto Cleanup;
	}

	//
	// Once the device is created successfully, framework frees the
	// DeviceInit memory and sets the pDeviceInit to NULL. So don't
	// call any WdfDeviceInit functions after that.
	//
	// Get the device context.
	//
	p_pdo = PdoGetData(hChild);
	p_fdo = FdoGetData(Device);

	p_pdo->p_fdo = p_fdo;
	p_pdo->SerialNo = SerialNo;
	p_pdo->PdoDevice = hChild;

	//
	// Set PnP properties for the child device.
	//
	WDF_DEVICE_PNP_CAPABILITIES_INIT(&pnpCaps);
	pnpCaps.Address  = SerialNo;
	pnpCaps.UINumber = SerialNo;

	WdfDeviceSetPnpCapabilities(hChild, &pnpCaps);

	//
	// Set Power properties for the child device
	//
	WDF_DEVICE_POWER_CAPABILITIES_INIT(&powerCaps);

	powerCaps.DeviceD1 = WdfTrue;
	powerCaps.WakeFromD1 = WdfTrue;
	powerCaps.DeviceWake = PowerDeviceD1;

	powerCaps.DeviceState[PowerSystemWorking]   = PowerDeviceD0;
	powerCaps.DeviceState[PowerSystemSleeping1] = PowerDeviceD1;
	powerCaps.DeviceState[PowerSystemSleeping2] = PowerDeviceD3;
	powerCaps.DeviceState[PowerSystemSleeping3] = PowerDeviceD3;
	powerCaps.DeviceState[PowerSystemHibernate] = PowerDeviceD3;
	powerCaps.DeviceState[PowerSystemShutdown] = PowerDeviceD3;

	WdfDeviceSetPowerCapabilities(hChild, &powerCaps);

	//
	// Set bus IB interface
	//
	p_fdo->bus_ib_ifc.port_id = (u8) SerialNo;
	p_fdo->bus_ib_ifc.pVipBusIfc = &p_fdo->mtnic_Ifc;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_pd_alloc = mlx4_pd_alloc;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_pd_free = mlx4_pd_free;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_uar_alloc = mlx4_uar_alloc;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_uar_free = mlx4_uar_free;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_mr_alloc = mlx4_mr_alloc;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_mr_free = mlx4_mr_free;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_mr_enable = mlx4_mr_enable;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_qp_get_region = mlx4_qp_get_region;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_alloc_hwq_res = mlx4_alloc_hwq_res;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_free_hwq_res = mlx4_free_hwq_res;

	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_cq_alloc = mlx4_cq_alloc;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_cq_free = mlx4_cq_free;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_cq_modify = mlx4_cq_modify;
	
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_register_mac = mlx4_register_mac;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_unregister_mac = mlx4_unregister_mac;
	
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_srq_alloc = mlx4_srq_alloc;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_srq_free = mlx4_srq_free;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_srq_invalidate = mlx4_srq_invalidate;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_srq_remove = mlx4_srq_remove;

	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_qp_alloc = mlx4_qp_alloc;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_qp_free = mlx4_qp_free;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_qp_remove = mlx4_qp_remove;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_qp_modify = mlx4_qp_modify;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_qp_to_ready = mlx4_qp_to_ready;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_qp_reserve_range = mlx4_qp_reserve_range;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_qp_release_range = mlx4_qp_release_range;

	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_alloc_cmd_mailbox = mlx4_alloc_cmd_mailbox;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_free_cmd_mailbox = mlx4_free_cmd_mailbox;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_cmd = imlx4_cmd;

	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_INIT_PORT = mlx4_INIT_PORT;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_CLOSE_PORT = mlx4_CLOSE_PORT;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_add_eq = mlx4_add_eq;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_remove_eq = mlx4_remove_eq;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_register_ev_cb = mlx4_reset_cb_register;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_unregister_ev_cb = mlx4_reset_cb_unregister;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_reset_request = mlx4_reset_request;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_reset_execute = mlx4_reset_execute;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_reset_ready = mlx4_reset_ready;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_register_vlan = mlx4_register_vlan;
	p_fdo->bus_ib_ifc.mlx4_interface.mlx4_unregister_vlan = mlx4_unregister_vlan;
	
	//
	// Create a custom interface so that other drivers can
	// query (IRP_MN_QUERY_INTERFACE) and use our callbacks directly.
	//
	p_fdo->bus_ib_ifc.i.Context = p_pdo;

	WDF_QUERY_INTERFACE_CONFIG_INIT( &p_pdo->qiMlx4Bus,
		(PINTERFACE) &p_fdo->bus_ib_ifc,
		&MLX4_BUS_IB_INTERFACE_GUID, EvtDeviceProcessQueryInterfaceRequest);

	status = WdfDeviceAddQueryInterface( hChild, &p_pdo->qiMlx4Bus );
	if (!NT_SUCCESS(status))
		goto Cleanup;

	//
	// Expose also PCI.SYS interface for MLX4_HCA
	//
	WDF_QUERY_INTERFACE_CONFIG_INIT( &p_pdo->qiPciBus,
		(PINTERFACE) &p_fdo->pci_dev.bus_pci_ifc,
		&GUID_BUS_INTERFACE_STANDARD, NULL);

	// TODO: Soft Reset - how tobar getting interface during RESET_IN_PROGRESS
	// maybe - using EvtDeviceProcessQueryInterfaceRequest
	status = WdfDeviceAddQueryInterface( hChild, &p_pdo->qiPciBus );
	if (!NT_SUCCESS(status))
		goto Cleanup;

	//
	// Add this device to the FDO's collection of children.
	// After the child device is added to the static collection successfully,
	// driver must call WdfPdoMarkMissing to get the device deleted. It
	// shouldn't delete the child device directly by calling WdfObjectDelete.
	//
	status = WdfFdoAddStaticChild(Device, hChild);
	if (!NT_SUCCESS(status)) {
		goto Cleanup;
	}

	return status;

Cleanup:
	KdPrint(("BusEnum: Bus_CreatePdo failed %x\n", status));

	//
	// Call WdfDeviceInitFree if you encounter an error before the
	// device is created. Once the device is created, framework
	// NULLs the pDeviceInit value.
	//
	if (pDeviceInit != NULL) {
		WdfDeviceInitFree(pDeviceInit);
	}

	if(hChild) {
		WdfObjectDelete(hChild);
	}

	return status;
}


