/*
 * Copyright (c) 2005 Mellanox Technologies. All rights reserved.
 * Copyright (c) 2005, 2006, 2007 Cisco Systems, Inc. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include "mlx4.h"
#include "cmd.h"
#include "fw.h"
#include "eq.h"

/*
 * Must be packed because start is 64 bits but only aligned to 32 bits.
 */
struct mlx4_eq_context {
	__be32			flags;
	u16			reserved1[3];
	__be16			page_offset;
	u8			log_eq_size;
	u8			reserved2[4];
	u8			eq_period;
	u8			reserved3;
	u8			eq_max_count;
	u8			reserved4[3];
	u8			intr;
	u8			log_page_size;
	u8			reserved5[2];
	u8			mtt_base_addr_h;
	__be32			mtt_base_addr_l;
	u32			reserved6[2];
	__be32			consumer_index;
	__be32			producer_index;
	u32			reserved7[4];
};

#define MLX4_EQ_STATUS_OK	   ( 0 << 28)
#define MLX4_EQ_STATUS_WRITE_FAIL  (10 << 28)
#define MLX4_EQ_OWNER_SW	   ( 0 << 24)
#define MLX4_EQ_OWNER_HW	   ( 1 << 24)
#define MLX4_EQ_FLAG_EC		   ( 1 << 18)
#define MLX4_EQ_FLAG_OI		   ( 1 << 17)
#define MLX4_EQ_STATE_ARMED	   ( 9 <<  8)
#define MLX4_EQ_STATE_FIRED	   (10 <<  8)
#define MLX4_EQ_STATE_ALWAYS_ARMED (11 <<  8)

#define MLX4_ASYNC_EVENT_MASK ((1ull << MLX4_EVENT_TYPE_PATH_MIG)	    | \
			       (1ull << MLX4_EVENT_TYPE_COMM_EST)	    | \
			       (1ull << MLX4_EVENT_TYPE_SQ_DRAINED)	    | \
			       (1ull << MLX4_EVENT_TYPE_CQ_ERROR)	    | \
			       (1ull << MLX4_EVENT_TYPE_WQ_CATAS_ERROR)	    | \
			       (1ull << MLX4_EVENT_TYPE_EEC_CATAS_ERROR)    | \
			       (1ull << MLX4_EVENT_TYPE_PATH_MIG_FAILED)    | \
			       (1ull << MLX4_EVENT_TYPE_WQ_INVAL_REQ_ERROR) | \
			       (1ull << MLX4_EVENT_TYPE_WQ_ACCESS_ERROR)    | \
			       (1ull << MLX4_EVENT_TYPE_PORT_CHANGE)	    | \
			       (1ull << MLX4_EVENT_TYPE_ECC_DETECT)	    | \
			       (1ull << MLX4_EVENT_TYPE_SRQ_CATAS_ERROR)    | \
			       (1ull << MLX4_EVENT_TYPE_SRQ_QP_LAST_WQE)    | \
			       (1ull << MLX4_EVENT_TYPE_SRQ_LIMIT)	    | \
			       (1ull << MLX4_EVENT_TYPE_CMD))

#pragma warning( disable : 4706)
static int mlx4_eq_int(struct mlx4_dev *dev, struct mlx4_eq *eq)
{
	struct mlx4_eqe *eqe;
	int cqn;
	int eqes_found = 0;
	int set_ci = 0;
	static const uint32_t cDpcMaxTime = 10000; //max time to spend in a while loop
		
    uint64_t start = cl_get_time_stamp();

	while ((eqe = next_eqe_sw(eq))) {
		/*
		 * Make sure we read EQ entry contents after we've
		 * checked the ownership bit.
		 */
		rmb();

		switch (eqe->type) {
		case MLX4_EVENT_TYPE_COMP:
			cqn = be32_to_cpu(eqe->event.comp.cqn) & 0xffffff;
			mlx4_cq_completion(dev, cqn);
			break;

		case MLX4_EVENT_TYPE_PATH_MIG:
		case MLX4_EVENT_TYPE_COMM_EST:
		case MLX4_EVENT_TYPE_SQ_DRAINED:
		case MLX4_EVENT_TYPE_SRQ_QP_LAST_WQE:
		case MLX4_EVENT_TYPE_WQ_CATAS_ERROR:
		case MLX4_EVENT_TYPE_PATH_MIG_FAILED:
		case MLX4_EVENT_TYPE_WQ_INVAL_REQ_ERROR:
		case MLX4_EVENT_TYPE_WQ_ACCESS_ERROR:
			mlx4_qp_event(dev, be32_to_cpu(eqe->event.qp.qpn) & 0xffffff,
				      eqe->type);
			break;

		case MLX4_EVENT_TYPE_SRQ_LIMIT:
		case MLX4_EVENT_TYPE_SRQ_CATAS_ERROR:
			mlx4_srq_event(dev, be32_to_cpu(eqe->event.srq.srqn) & 0xffffff,
				      eqe->type);
			break;

		case MLX4_EVENT_TYPE_CMD:
			mlx4_cmd_event(dev,
				       be16_to_cpu(eqe->event.cmd.token),
				       eqe->event.cmd.status,
				       be64_to_cpu(eqe->event.cmd.out_param));
			break;

		case MLX4_EVENT_TYPE_PORT_CHANGE:
			mlx4_dispatch_event(dev, eqe->type, eqe->subtype,
					    be32_to_cpu(eqe->event.port_change.port) >> 28);
			break;

		case MLX4_EVENT_TYPE_CQ_ERROR:
			mlx4_warn(dev, "CQ %s on CQN %06x\n",
				  eqe->event.cq_err.syndrome == 1 ?
				  "overrun" : "access violation",
				  be32_to_cpu(eqe->event.cq_err.cqn) & 0xffffff);
			mlx4_cq_event(dev, be32_to_cpu(eqe->event.cq_err.cqn),
				      eqe->type);
			break;

		case MLX4_EVENT_TYPE_EQ_OVERFLOW:
			mlx4_warn(dev, "EQ overrun on EQN %d\n", eq->eqn);
			break;

		case MLX4_EVENT_TYPE_EEC_CATAS_ERROR:
		case MLX4_EVENT_TYPE_ECC_DETECT:
		default:
			mlx4_warn(dev, "Unhandled event %02x(%02x) on EQ %d at index %u\n",
				  eqe->type, eqe->subtype, eq->eqn, eq->cons_index);
			
			break;
		};

		++eq->cons_index;
		eqes_found = 1;
		++set_ci;

		/*
		 * The HCA will think the queue has overflowed if we
		 * don't tell it we've been processing events.  We
		 * create our EQs with MLX4_NUM_SPARE_EQE extra
		 * entries, so we must update our consumer index at
		 * least that often.
		 */
		if (unlikely(set_ci >= MLX4_NUM_SPARE_EQE)) {
			/*
			 * Conditional on hca_type is OK here because
			 * this is a rare case, not the fast path.
			 */
			eq_set_ci(eq, 0);
			set_ci = 0;
		}
		
		if (cl_get_time_stamp() - start > cDpcMaxTime ) {
			break; //allow other DPCs as well
		}
	}

	eq_set_ci(eq, 1);

	return eqes_found;
}
#pragma warning(disable:4706)

static void mlx4_dpc( PRKDPC dpc, 
	PVOID ctx, PVOID arg1, PVOID arg2 )
{
	struct mlx4_eq *eq  = ctx;

	UNREFERENCED_PARAMETER(dpc);
	UNREFERENCED_PARAMETER(arg1);
	UNREFERENCED_PARAMETER(arg2);

	spin_lock_dpc(&eq->lock);
	mlx4_eq_int(eq->dev, eq);
	spin_unlock_dpc(&eq->lock);
}

static BOOLEAN legacy_isr(struct mlx4_dev *dev)
{
	int i;
	int work = 0;
	struct mlx4_priv *priv = mlx4_priv(dev);

	for (i = 0; i < MLX4_NUM_EQ; ++i) {
		if ( next_eqe_sw(&priv->eq_table.eq[i]) ) {
			work = 1;
			/* another interrupt may happen instantly after writel above.
			If it comes to another processor, mlx4_interrupt will be called
			and try to schedule the same DPC. So we protect KeInsertQueueDpc
			from that race */
			
			while(InterlockedCompareExchange(&dev->pdev->dpc_lock, 1, 0));
			
			KeInsertQueueDpc(&priv->eq_table.eq[i].dpc, NULL, NULL);
			InterlockedCompareExchange(&dev->pdev->dpc_lock, 0, 1);
		} else {
			/* re-arm the EQ for a case when interrupt comes before EQE
			and we didn't scheduled the DPC */
			eq_set_ci(&priv->eq_table.eq[i], 1);
		}
	}

	for (i = MLX4_NUM_EQ; i <= priv->eq_table.max_extra_eqs; ++i) {
		if (priv->eq_table.eq[i].isr) {
			int ret = 0;
			if ( next_eqe_sw(&priv->eq_table.eq[i]) ) {
				ret = priv->eq_table.eq[i].isr(priv->eq_table.eq[i].ctx);
				work |= ret;
			} else {
				eq_set_ci(&priv->eq_table.eq[i], 1);
			}
		}
	}

	return (BOOLEAN)work;
}

static BOOLEAN mlx4_interrupt(
	IN struct _KINTERRUPT *Interrupt,
	IN PVOID ServiceContext
	)
{
	struct mlx4_dev *dev = ServiceContext;
	struct mlx4_priv *priv = mlx4_priv(dev);

	UNUSED_PARAM(Interrupt);
	writel(priv->eq_table.clr_mask, priv->eq_table.clr_int);
	return legacy_isr(dev);
}

static void mlx4_dpc_msix( PRKDPC dpc, 
	PVOID ctx, PVOID arg1, PVOID arg2 )
{
	struct mlx4_eq *eq  = ctx;

	UNREFERENCED_PARAMETER(dpc);
	UNREFERENCED_PARAMETER(arg1);
	UNREFERENCED_PARAMETER(arg2);

	mlx4_eq_int(eq->dev, eq);
}


#if 1//WORKAROUND_POLL_EQ

BOOLEAN
IsrSynchronizeRoutine(
	IN PVOID  SynchronizeContext
	)
{
	struct mlx4_dev *dev = (struct mlx4_dev *)SynchronizeContext;
    
	mlx4_interrupt(dev->pdev->int_obj,dev);
    
	return TRUE;
}


VOID eq_polling_thread(void *ctx) 
{
#define POLLING_INTERVAL_MS	50
	NTSTATUS status;
	struct mlx4_priv *priv = (struct mlx4_priv *)ctx;
	PVOID wait_objects[2];
	LARGE_INTEGER  wait_time;

	wait_objects[0] = &priv->eq_table.thread_stop_event;
	wait_objects[1] = &priv->eq_table.thread_start_event;

	for(;;){

		/* before start polling */
		for (;;) {
			status = KeWaitForMultipleObjects( 2, wait_objects, 
											   WaitAny, Executive, KernelMode, FALSE, NULL, NULL );

			if ( status == STATUS_WAIT_0 ){/* thread stopped */
				break;		
			}

			/* start polling */
			if ( status == STATUS_WAIT_1 ){
				break;		
			}

		}

		if(priv->eq_table.bTerminated) break;
		if ( status == STATUS_WAIT_0 ) continue;/* thread stopped, wait for start again */

		/* polling */
		wait_time.QuadPart = -(int64_t)(((uint64_t)POLLING_INTERVAL_MS) * 10000);
		for (;;) {
			//mlx4_interrupt( NULL, &priv->dev );
			KeSynchronizeExecution(priv->dev.pdev->int_obj, IsrSynchronizeRoutine, &priv->dev);

			status = KeWaitForSingleObject( &priv->eq_table.thread_stop_event, 
											Executive, KernelMode, FALSE, &wait_time );
			if ( status == STATUS_SUCCESS ) {
				//KeClearEvent(&priv->eq_table.thread_stop_event);
				break;		/* thread stopped */
			}
		}

		if(priv->eq_table.bTerminated) break;
	}

	PsTerminateSystemThread(STATUS_SUCCESS);

}


void mlx4_poll_eq(struct ib_device *device, BOOLEAN bStart)
{
	LONG signalled=0;
	struct mlx4_priv *priv = mlx4_priv(device->dma_device);

	if(bStart){
		/* signal start of polling */
		signalled = KeSetEvent( 	
			&priv->eq_table.thread_start_event, IO_NO_INCREMENT, FALSE );
	}else{
		/* signal end of polling */
		signalled = KeSetEvent( 	
			&priv->eq_table.thread_stop_event, IO_NO_INCREMENT, FALSE );
	}

}

#endif


BOOLEAN
mlx4_msi_x_interrupt (
	IN PKINTERRUPT  Interrupt,
	IN PVOID  ServiceContext,
	IN ULONG  MessageId 
	)
{
	int eq_ix;
	struct mlx4_eq *eq;
	struct mlx4_priv *priv;
	struct mlx4_dev *dev = ServiceContext;

	UNUSED_PARAM(Interrupt);
	
	// Check, that it is our interrupt
	if ( dev->signature != MLX4_DEV_SIGNATURE)
		return FALSE;

	if (dev->flags & MLX4_FLAG_MSI_X) {
		// MSI-X mode
		priv = mlx4_priv(dev);
		eq_ix = dev->pdev->p_msix_map[MessageId].eq_ix;
		eq = &priv->eq_table.eq[eq_ix];
		if (eq_ix >= MLX4_NUM_EQ) {
			if (eq->isr) 
				eq->isr(eq->ctx);
			else 
				eq_set_ci(eq, 1);
		}
		else
			KeInsertQueueDpc(&eq->dpc, NULL, NULL);
		return TRUE;
	}

	// legacy mode
	return mlx4_interrupt(NULL,dev);
}

static int mlx4_MAP_EQ(struct mlx4_dev *dev, u64 event_mask, int unmap,
			int eq_num)
{
	return mlx4_cmd(dev, event_mask, (unmap << 31) | eq_num,
			0, MLX4_CMD_MAP_EQ, MLX4_CMD_TIME_CLASS_B);
}

static int mlx4_SW2HW_EQ(struct mlx4_dev *dev, struct mlx4_cmd_mailbox *mailbox,
			 int eq_num)
{
	return mlx4_cmd(dev, mailbox->dma.da, eq_num, 0, MLX4_CMD_SW2HW_EQ,
			MLX4_CMD_TIME_CLASS_A);
}

static int mlx4_HW2SW_EQ(struct mlx4_dev *dev, struct mlx4_cmd_mailbox *mailbox,
			 int eq_num)
{
	return mlx4_cmd_box(dev, 0, mailbox->dma.da, eq_num, 0, MLX4_CMD_HW2SW_EQ,
			    MLX4_CMD_TIME_CLASS_A);
}

static void __iomem *mlx4_get_eq_uar(struct mlx4_dev *dev, struct mlx4_eq *eq)
{
	struct mlx4_priv *priv = mlx4_priv(dev);
	int index;

	index = eq->eqn / 4 - dev->caps.reserved_eqs / 4;

	if (!priv->eq_table.uar_map[index]) {
		priv->eq_table.uar_map[index] =
			ioremap(pci_resource_start(dev->pdev, 2) +
				((eq->eqn / 4) << PAGE_SHIFT),
				PAGE_SIZE, MmNonCached);
		if (!priv->eq_table.uar_map[index]) {
			mlx4_err(dev, "Couldn't map EQ doorbell for EQN 0x%06x\n",
				 eq->eqn);
			return NULL;
		}
	}

	return priv->eq_table.uar_map[index] + 0x800 + 8 * (eq->eqn % 4);
}

/* for legacy interrupts the function returns 0 */
/* for MSI-X - one vector per EQ */
/* if cpu=0, it selects the first free vector */
/* otherwise it returns free vector, allocated to this cpu */
/* if cpu=0 and all vectors are taken it returns the first vector */
/* if cpu!=0 and free vector for that cpu was not found - returns MLX4_NUM_UNKNOWN */
int mlx4_alloc_msi_irqn(struct mlx4_dev *dev, KAFFINITY cpu, int eq_ix )
{
	int i;
	struct pci_dev *pdev = dev->pdev;

	if (!(dev->flags & MLX4_FLAG_MSI_X)) {
		if (pdev->p_msix_map) {
			pdev->p_msix_map[0].ref_cnt++;
			pdev->p_msix_map[0].eq_ix = eq_ix;
		}
		return 0;
	}
		
	for (i=0; i < pdev->n_msi_vectors_alloc; ++i) {
		if (pdev->p_msix_map[i].ref_cnt)
			continue;
		if (!cpu || (pdev->p_msix_map[i].cpu & cpu)) {
			pdev->p_msix_map[i].ref_cnt++;
			pdev->p_msix_map[i].eq_ix = eq_ix;
			return i;
		}
	}

	/* all vectors are taken - return error  */
    ASSERT(0);
	return MLX4_NUM_UNKNOWN;
}

void mlx4_free_msi_irqn(struct mlx4_dev *dev, int intr)
{
	if (dev->pdev->p_msix_map) {
		dev->pdev->p_msix_map[intr].ref_cnt--;
		ASSERT(dev->pdev->p_msix_map[intr].ref_cnt>=0);
	}
}

static int mlx4_create_eq(struct mlx4_dev *dev, int nent,
		KAFFINITY cpu, int eq_ix, struct mlx4_eq *eq)
{
	struct mlx4_priv *priv = mlx4_priv(dev);
	struct mlx4_cmd_mailbox *mailbox;
	struct mlx4_eq_context *eq_context;
	int npages;
	u64 *dma_list = NULL;
	dma_addr_t t;
	u64 mtt_addr;
	int err = -ENOMEM;
	int i;
	int intr;

	eq->valid = 0;
	intr = mlx4_alloc_msi_irqn(dev, cpu, eq_ix);
	if (intr == MLX4_NUM_UNKNOWN) {
		err = -ENOENT;
		goto err_out;
	}
	eq->dev   = dev;
	eq->irq = (u16)intr;
	eq->nent  = roundup_pow_of_two(max(nent, 2));
	eq->cpu = cpu;
	npages = (int)(NEXT_PAGE_ALIGN(eq->nent * MLX4_EQ_ENTRY_SIZE) / PAGE_SIZE);

	eq->page_list = kmalloc(npages * sizeof *eq->page_list,
				GFP_KERNEL);
	if (!eq->page_list)
		goto err_malloc;

	for (i = 0; i < npages; ++i)
		eq->page_list[i].buf = NULL;

	dma_list = kmalloc(npages * sizeof *dma_list, GFP_KERNEL);
	if (!dma_list)
		goto err_out_free;

	mailbox = mlx4_alloc_cmd_mailbox(dev);
	if (IS_ERR(mailbox))
		goto err_out_free;
	eq_context = (struct mlx4_eq_context *)mailbox->buf;

	for (i = 0; i < npages; ++i) {
		eq->page_list[i].buf = dma_alloc_coherent(&dev->pdev->dev,
							  PAGE_SIZE, &t, GFP_KERNEL);
		if (!eq->page_list[i].buf)
			goto err_out_free_pages;

		dma_list[i] = t.da;
		eq->page_list[i].map = t;

		memset(eq->page_list[i].buf, 0, PAGE_SIZE);
	}

	eq->eqn = mlx4_bitmap_alloc(&priv->eq_table.bitmap);
	if (eq->eqn == -1)
		goto err_out_free_pages;

	eq->doorbell = mlx4_get_eq_uar(dev, eq);
	if (!eq->doorbell) {
		err = -ENOMEM;
		goto err_out_free_eq;
	}

	err = mlx4_mtt_init(dev, npages, PAGE_SHIFT, &eq->mtt);
	if (err)
		goto err_out_free_eq;

	err = mlx4_write_mtt(dev, &eq->mtt, 0, npages, dma_list);
	if (err)
		goto err_out_free_mtt;

	memset(eq_context, 0, sizeof *eq_context);
	eq_context->flags	  = cpu_to_be32(MLX4_EQ_STATUS_OK   |
						MLX4_EQ_STATE_ARMED);
	eq_context->log_eq_size	  = (u8)ilog2(eq->nent);
	eq_context->intr	  = (u8)intr;
	eq_context->log_page_size = PAGE_SHIFT - MLX4_ICM_PAGE_SHIFT;

	mtt_addr = mlx4_mtt_addr(dev, &eq->mtt);
	eq_context->mtt_base_addr_h = (u8)(mtt_addr >> 32);
	eq_context->mtt_base_addr_l = cpu_to_be32(mtt_addr & 0xffffffff);

	err = mlx4_SW2HW_EQ(dev, mailbox, eq->eqn);
	if (err) {
		mlx4_warn(dev, "SW2HW_EQ failed (%d)\n", err);
		goto err_out_free_mtt;
	}

	mlx4_dbg(dev, "EQ created: EQN %d, size %d(%d), MSI ID %d, CPU %#x\n", 
		eq->eqn, nent, eq->nent, intr, cpu );

	kfree(dma_list);
	mlx4_free_cmd_mailbox(dev, mailbox);

	eq->cons_index = 0;
	eq->valid = 1;

	return 0;

err_out_free_mtt:
	mlx4_mtt_cleanup(dev, &eq->mtt);

err_out_free_eq:
	mlx4_bitmap_free(&priv->eq_table.bitmap, eq->eqn);

err_out_free_pages:
	for (i = 0; i < npages; ++i)
		if (eq->page_list[i].buf)
			dma_free_coherent(&dev->pdev->dev, PAGE_SIZE,
					  eq->page_list[i].buf,
					  eq->page_list[i].map);

	mlx4_free_cmd_mailbox(dev, mailbox);

err_out_free:
	kfree(eq->page_list);
	kfree(dma_list);

err_malloc:
	mlx4_free_msi_irqn(dev, eq->irq);

err_out:
	return err;
}

static void mlx4_free_eq(struct mlx4_dev *dev,
			 struct mlx4_eq *eq)
{
	struct mlx4_priv *priv = mlx4_priv(dev);
	struct mlx4_cmd_mailbox *mailbox;
	int err;
	int npages = (int)(NEXT_PAGE_ALIGN(MLX4_EQ_ENTRY_SIZE * eq->nent) / PAGE_SIZE);
	int i;

	if (!eq->valid)
		return;
	eq->valid = 0;
	
	mailbox = mlx4_alloc_cmd_mailbox(dev);
	if (IS_ERR(mailbox)) 
		mailbox = NULL;
	else {
		err = mlx4_HW2SW_EQ(dev, mailbox, eq->eqn);
		if (err)
			mlx4_warn(dev, "HW2SW_EQ failed (%d)\n", err);
	}

#if 0
	{
		mlx4_dbg(dev, "Dumping EQ context %02x:\n", eq->eqn);
		for (i = 0; i < sizeof (struct mlx4_eq_context) / 4; ++i) {
			if (i % 4 == 0)
				printk("[%02x] ", i * 4);
			printk(" %08x", be32_to_cpup(mailbox->buf + i * 4));
			if ((i + 1) % 4 == 0)
				printk("\n");
		}
	}
#endif	

	mlx4_mtt_cleanup(dev, &eq->mtt);
	for (i = 0; i < npages; ++i)
		pci_free_consistent(dev->pdev, PAGE_SIZE,
				    eq->page_list[i].buf,
				    eq->page_list[i].map);

	kfree(eq->page_list);
	mlx4_bitmap_free(&priv->eq_table.bitmap, eq->eqn);
	mlx4_free_cmd_mailbox(dev, mailbox);
	mlx4_free_msi_irqn(dev, eq->irq);
}

static void mlx4_free_irqs(struct mlx4_dev *dev)
{
	struct mlx4_eq_table *eq_table = &mlx4_priv(dev)->eq_table;

	if (eq_table->have_irq)
		free_irq(dev);
}

static int mlx4_map_clr_int(struct mlx4_dev *dev)
{
	struct mlx4_priv *priv = mlx4_priv(dev);

	priv->clr_base = ioremap(pci_resource_start(dev->pdev, priv->fw.clr_int_bar) +
				 priv->fw.clr_int_base, MLX4_CLR_INT_SIZE, MmNonCached);
	if (!priv->clr_base) {
		mlx4_err(dev, "Couldn't map interrupt clear register, aborting.\n");
		return -ENOMEM;
	}

	return 0;
}

static void mlx4_unmap_clr_int(struct mlx4_dev *dev)
{
	struct mlx4_priv *priv = mlx4_priv(dev);

	if (priv->clr_base) 
		iounmap(priv->clr_base, MLX4_CLR_INT_SIZE);
}

int mlx4_init_eq_table(struct mlx4_dev *dev)
{
	struct mlx4_priv *priv = mlx4_priv(dev);
	int err;
	int i;

	err = mlx4_bitmap_init(&priv->eq_table.bitmap, dev->caps.num_eqs,
			       dev->caps.num_eqs - 1, dev->caps.reserved_eqs);
	if (err)
		return err;

	for (i = 0; i < ARRAY_SIZE(priv->eq_table.uar_map); ++i)
		priv->eq_table.uar_map[i] = NULL;

	err = mlx4_map_clr_int(dev);
	if (err)
		goto err_out_free;

	priv->eq_table.clr_mask =
		swab32(1 << (priv->eq_table.inta_pin & 31));
	priv->eq_table.clr_int  = priv->clr_base +
		(priv->eq_table.inta_pin < 32 ? 4 : 0);

	/* EQ creation should be in this order because of MSI-X vector allocation way */
	err = mlx4_create_eq(dev, MLX4_NUM_ASYNC_EQE + MLX4_NUM_SPARE_EQE,
		0, MLX4_EQ_ASYNC, &priv->eq_table.eq[MLX4_EQ_ASYNC]);
	if (err)
		goto err_out_unmap;

	err = mlx4_create_eq(dev, dev->caps.num_cqs + MLX4_NUM_SPARE_EQE,
		0, MLX4_EQ_COMP, &priv->eq_table.eq[MLX4_EQ_COMP]);
	if (err)
		goto err_out_comp;

	/* init DPC stuff */
	dev->pdev->dpc_lock = 0;
	for (i = 0; i < MLX4_NUM_EQS; ++i) {
		spin_lock_init( &priv->eq_table.eq[i].lock );	
		KeInitializeDpc( &priv->eq_table.eq[i].dpc, 
			(dev->flags & MLX4_FLAG_MSI_X) ? mlx4_dpc_msix : mlx4_dpc, 
			&priv->eq_table.eq[i]);
		priv->eq_table.eq[i].eq_ix = i;
	}

	// connect interrupts
#ifdef USE_WDM_INTERRUPTS
	err = request_irq( dev, 
		mlx4_interrupt, dev, 
		mlx4_msi_x_interrupt,
		&dev->pdev->int_obj );
	if (err)
		goto err_out_async;
#else
	// not implemented - prevent compilation
	#error Interrupts in WDF style are not implemented !
	goto err_out_async;
#endif		
	priv->eq_table.have_irq = 1;

	err = mlx4_MAP_EQ(dev, MLX4_ASYNC_EVENT_MASK, 0,
			  priv->eq_table.eq[MLX4_EQ_ASYNC].eqn);
	if (err) {
		mlx4_warn(dev, "MAP_EQ for async EQ %d failed (%d)\n",
			   priv->eq_table.eq[MLX4_EQ_ASYNC].eqn, err);
		goto err_map_eq;
	}

	mlx4_dbg(dev, "Commands and async events are mapped to EQ %d\n",
		priv->eq_table.eq[MLX4_EQ_ASYNC].eqn);

	for (i = 0; i < MLX4_NUM_EQ; ++i)
		eq_set_ci(&priv->eq_table.eq[i], 1);

#if 1//WORKAROUND_POLL_EQ
	{ /* Create a thread for polling EQs in case of missing interrupts from the card  */
		NTSTATUS status;
		OBJECT_ATTRIBUTES   attr;
		HANDLE  handle;

		KeInitializeEvent(&priv->eq_table.thread_start_event, SynchronizationEvent, FALSE);
		KeInitializeEvent(&priv->eq_table.thread_stop_event, SynchronizationEvent, FALSE);
		priv->eq_table.bTerminated = FALSE;
		priv->eq_table.threadObject = NULL;
		InitializeObjectAttributes( &attr, NULL, OBJ_KERNEL_HANDLE, NULL, NULL );
		status = PsCreateSystemThread( &handle, 
			THREAD_ALL_ACCESS, &attr, NULL, NULL, eq_polling_thread, priv );
		if (NT_SUCCESS(status)) {
			status = ObReferenceObjectByHandle(
				handle,
				THREAD_ALL_ACCESS,
				NULL,
				KernelMode,
				&priv->eq_table.threadObject,
				NULL
				);

			ASSERT(status == STATUS_SUCCESS); //

			status = ZwClose(handle);

			ASSERT(NT_SUCCESS(status)); // Should always succeed

		}
	}
#endif

	return 0;

err_map_eq:
	mlx4_free_irqs(dev);

err_out_async:
	mlx4_free_eq(dev, &priv->eq_table.eq[MLX4_EQ_COMP]);

err_out_comp:
	mlx4_free_eq(dev, &priv->eq_table.eq[MLX4_EQ_ASYNC]);

err_out_unmap:
	mlx4_unmap_clr_int(dev);

err_out_free:
	mlx4_bitmap_cleanup(&priv->eq_table.bitmap);
	return err;
}

int mlx4_add_eq(struct mlx4_dev *dev, int nent,
	KAFFINITY cpu, PISR_FUNC func, PVOID func_context ,
	u8* p_eq_num, struct mlx4_eq ** p_eq)
{
	struct mlx4_priv *priv = mlx4_priv(dev);
	int err;
	u8 i, new_eq = 0;

	if ( mlx4_is_barred(dev) )
		return -EFAULT;
	
	for (i = MLX4_NUM_EQ; i < MLX4_NUM_EQS; i++) {
		if(priv->eq_table.eq[i].isr == NULL) {
			new_eq = i;
			break;
		}
	}
	if (new_eq == 0)
		return -ENOMEM;

	err = mlx4_create_eq(dev, nent,
		cpu, new_eq, &priv->eq_table.eq[new_eq]);
	if (err) {
		return err;
	}

	*p_eq = &priv->eq_table.eq[new_eq ];
	*p_eq_num = new_eq;
	priv->eq_table.eq[new_eq].isr = func;
	priv->eq_table.eq[new_eq].ctx = func_context;
	priv->eq_table.max_extra_eqs = max(priv->eq_table.max_extra_eqs, new_eq);
	return 0;
}

void mlx4_remove_eq(struct mlx4_dev *dev, u8 eq_num)
{
	struct mlx4_priv *priv = mlx4_priv(dev);
	struct mlx4_eq_table *eq_table = &mlx4_priv(dev)->eq_table;

	priv->eq_table.eq[eq_num].isr = NULL;
	priv->eq_table.eq[eq_num].ctx = NULL;

	if (priv->eq_table.max_extra_eqs == eq_num)
		priv->eq_table.max_extra_eqs--;

	mlx4_free_eq(dev, &priv->eq_table.eq[eq_num]);


	if (eq_table->have_irq) {
		int err;
		
		free_irq(dev);

	#ifdef USE_WDM_INTERRUPTS
		err = request_irq( dev, 
			mlx4_interrupt, dev, 
			mlx4_msi_x_interrupt,
			&dev->pdev->int_obj );
		if (err) {
			mlx4_err(dev, "request_irq failed (%d)\n",
				err);
			ASSERT(0);
		}
		else
			priv->eq_table.have_irq = 1;
	#else
		// not implemented - prevent compilation
		#error Interrupts in WDF style are not implemented !
	#endif		

	}
}

void mlx4_cleanup_eq_table(struct mlx4_dev *dev)
{
	struct mlx4_priv *priv = mlx4_priv(dev);
	int i;

#if 1//WORKAROUND_POLL_EQ
	/* stop the EQ polling thread */
	if (priv->eq_table.threadObject) { 
	#define WAIT_TIME_MS	3000
		NTSTATUS status;
		LARGE_INTEGER  wait_time;
		LONG signalled;

		priv->eq_table.bTerminated = TRUE;

		/* signal polling stopped in case it is not */
		signalled = KeSetEvent(
			&priv->eq_table.thread_stop_event, IO_NO_INCREMENT, FALSE );

		/* wait for completion of the thread */
		wait_time.QuadPart = -(int64_t)(((uint64_t)WAIT_TIME_MS) * 10000);
		status = KeWaitForSingleObject( priv->eq_table.threadObject,
			Executive, KernelMode, FALSE, &wait_time );
		ASSERT(status == STATUS_SUCCESS);

		ObDereferenceObject(priv->eq_table.threadObject);

		/* cleanup */
		priv->eq_table.threadObject = NULL;
	}
#endif
    
	mlx4_MAP_EQ(dev, MLX4_ASYNC_EVENT_MASK, 1,
		    priv->eq_table.eq[MLX4_EQ_ASYNC].eqn);

	mlx4_free_irqs(dev);
	mlx4_free_eq(dev, &priv->eq_table.eq[MLX4_EQ_COMP]);
	mlx4_free_eq(dev, &priv->eq_table.eq[MLX4_EQ_ASYNC]);
	mlx4_unmap_clr_int(dev);

	for (i = 0; i < ARRAY_SIZE(priv->eq_table.uar_map); ++i)
		if (priv->eq_table.uar_map[i])
			iounmap(priv->eq_table.uar_map[i],PAGE_SIZE);

	mlx4_bitmap_cleanup(&priv->eq_table.bitmap);
}
