/*
 * Copyright (c) 2007 Mellanox Technologies. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include "mlx4.h"
#include "cmd.h"
#include "public.h"

extern NTSTATUS __create_child();

void mlx4_init_mac_table(struct mlx4_dev *dev, u8 port)
{
	struct mlx4_mac_table *table = &mlx4_priv(dev)->port[port].mac_table;
	int i;

	sema_init(&table->mac_sem, 1);
	for (i = 0; i < MLX4_MAX_MAC_NUM; i++) {
		table->entries[i] = 0;
		table->refs[i] = 0;
	}
	table->max = 1 << dev->caps.log_num_macs;
	table->total = 0;
}

void mlx4_init_vlan_table(struct mlx4_dev *dev, u8 port)
{
	struct mlx4_vlan_table *table = &mlx4_priv(dev)->port[port].vlan_table;
	int i;

	sema_init(&table->vlan_sem, 1);
	for (i = 0; i < MLX4_MAX_VLAN_NUM; i++) {
		table->entries[i] = 0;
		table->refs[i] = 0;
	}
	table->max = 1 << dev->caps.log_num_vlans;
	if(table->max > MLX4_MAX_VLAN_NUM)
	{
		table->max = MLX4_MAX_VLAN_NUM;
	}
	table->total = 0;
}

static int mlx4_SET_PORT_mac_table(struct mlx4_dev *dev, u8 port,
				   __be64 *entries)
{
	struct mlx4_cmd_mailbox *mailbox;
	u32 in_mod;
	int err;

	mailbox = mlx4_alloc_cmd_mailbox(dev);
	if (IS_ERR(mailbox))
		return PTR_ERR(mailbox);

	memcpy(mailbox->buf, entries, MLX4_MAC_TABLE_SIZE);

	in_mod = MLX4_SET_PORT_MAC_TABLE << 8 | port;
	err = mlx4_cmd(dev, mailbox->dma.da, in_mod, 1, MLX4_CMD_SET_PORT,
		       MLX4_CMD_TIME_CLASS_B);

	mlx4_free_cmd_mailbox(dev, mailbox);
	return err;
}

static void mlx4_addrconf_ifid_eui48_win(u8 *eui, u64 mac)
{
    u8 *p = (u8*)&mac+2; //mac 6 bytes
	memcpy(eui, p, 3);
	memcpy(eui + 5, p + 3, 3);
	eui[3] = 0xFF;
	eui[4] = 0xFE;
	eui[0] ^= 2;
}


static int update_ipv6_gids_win(struct mlx4_dev *dev, int port, int clear, u64 mac)
{
	struct mlx4_cmd_mailbox *mailbox;
	union ib_gid *gids, *tmpgids;
	int err;

	tmpgids = kzalloc(128 * sizeof *gids, GFP_ATOMIC);
	if (!tmpgids)
		return -ENOMEM;

	if (!clear) {
		mlx4_addrconf_ifid_eui48_win(&tmpgids[0].raw[8], cpu_to_be64(mac));
		tmpgids[0].global.subnet_prefix = cpu_to_be64(0xfe80000000000000LL);
	}

	mailbox = mlx4_alloc_cmd_mailbox(dev);
	if (IS_ERR(mailbox)) {
		err = PTR_ERR(mailbox);
		goto out;
	}

	gids = mailbox->buf;
	memcpy(gids, tmpgids, 128 * sizeof *gids);

	err = mlx4_cmd(dev, mailbox->dma.da, MLX4_SET_PORT_GID_TABLE << 8 | port,
		       1, MLX4_CMD_SET_PORT, MLX4_CMD_TIME_CLASS_B);

	mlx4_free_cmd_mailbox(dev, mailbox);

out:
	kfree(tmpgids);
	return err;
}


int mlx4_register_mac(struct mlx4_dev *dev, u8 port, u64 mac, int *index)
{
	struct mlx4_mac_table *table =
	&mlx4_priv(dev)->port[port - 1].mac_table;
	int i, err = 0;
	int free = -1;
	u64 valid = 1;

	if ( mlx4_is_barred(dev) )
		return -EFAULT;

	mlx4_dbg(dev, "Registering mac : 0x%llx\n", mac);
	down(&table->mac_sem);
	for (i = 0; i < MLX4_MAX_MAC_NUM; i++) {
		if (free < 0 && !table->refs[i]) {
			free = i;
			continue;
		}

		if (mac == (MLX4_MAC_MASK & be64_to_cpu(table->entries[i]))) {
			/* Mac already registered, increase refernce count */
			*index = i;
			++table->refs[i];
			goto out;
		}
	}
	mlx4_dbg(dev, "Free mac index is %d\n", free);

	if (table->total == table->max || free < 0) {
		/* No free mac entries */
		err = -ENOSPC;
		goto out;
	}

	/* Register new MAC */
	table->refs[free] = 1;
	table->entries[free] = cpu_to_be64(mac | valid << MLX4_MAC_VALID_SHIFT);

	err = mlx4_SET_PORT_mac_table(dev, port, table->entries);
	if (unlikely(err)) {
		mlx4_err(dev, "Failed adding mac: 0x%llx\n", mac);
		table->refs[free] = 0;
		table->entries[free] = 0;
		goto out;
	}

	*index = free;
	++table->total;

	//update port guid with mac address
	update_ipv6_gids_win(dev, port, 0, mac);

#if 0

// TODO: Tzachid 9/12/2009 Need to think of a better way of how to create the LLE
// interface

   
	if(!InterlockedExchange(&dev->pdev->ib_hca_created, 1))
	{
    	NTSTATUS status = STATUS_SUCCESS;
		status = __create_child(dev->pdev->p_wdf_device, BUS_HARDWARE_IDS, BUS_HARDWARE_DESCRIPTION, 0 );
		if (!NT_SUCCESS(status)) {
			 mlx4_err(dev, "__create_child (ib)failed with 0x%x\n", status);
			 dev->pdev->ib_hca_created = FALSE;
		}
	}
#endif
out:
	up(&table->mac_sem);
	return err;
}
EXPORT_SYMBOL_GPL(mlx4_register_mac);

void mlx4_unregister_mac(struct mlx4_dev *dev, u8 port, int index)
{
	struct mlx4_mac_table *table =
	&mlx4_priv(dev)->port[port - 1].mac_table;

	down(&table->mac_sem);
	if (!table->refs[index]) {
		mlx4_warn(dev, "No mac entry for index %d\n", index);
		goto out;
	}
	if (--table->refs[index]) {
		mlx4_warn(dev, "Have more references for index %d,"
			  "no need to modify mac table\n", index);
		goto out;
	}
	table->entries[index] = 0;
	if ( !mlx4_is_barred(dev) )
		mlx4_SET_PORT_mac_table(dev, port, table->entries);
	--table->total;
out:
	up(&table->mac_sem);
}
EXPORT_SYMBOL_GPL(mlx4_unregister_mac);

static int mlx4_SET_PORT_vlan_table(struct mlx4_dev *dev, u8 port,
				    __be32 *entries)
{
	struct mlx4_cmd_mailbox *mailbox;
	u32 in_mod;
	int err;

	mailbox = mlx4_alloc_cmd_mailbox(dev);
	if (IS_ERR(mailbox))
		return PTR_ERR(mailbox);

	memcpy(mailbox->buf, entries, MLX4_VLAN_TABLE_SIZE);
	in_mod = MLX4_SET_PORT_VLAN_TABLE << 8 | port;
	err = mlx4_cmd(dev, mailbox->dma.da, in_mod, 1, MLX4_CMD_SET_PORT,
		       MLX4_CMD_TIME_CLASS_B);

	mlx4_free_cmd_mailbox(dev, mailbox);

	return err;
}

int mlx4_register_vlan(struct mlx4_dev *dev, u8 port, u16 vlan, int *index)
{
	struct mlx4_vlan_table *table =
	&mlx4_priv(dev)->port[port - 1].vlan_table;
	int i, err = 0;
	int free = -1;

	down(&table->vlan_sem);
	for (i = 0; i < MLX4_MAX_VLAN_NUM; i++) {
		if (free < 0 && (table->refs[i] == 0)) {
			free = i;
			continue;
		}

		if (table->refs[i] &&
		    (vlan == (MLX4_VLAN_MASK &
			      be32_to_cpu(table->entries[i])))) {
			/* Vlan already registered, increase refernce count */
			*index = i;
			++table->refs[i];
			goto out;
		}
	}

	if (table->total == table->max || free < 0) {
		/* No free vlan entries */
		err = -ENOSPC;
		goto out;
	}

	/* Register new MAC */
	table->refs[free] = 1;
	table->entries[free] = cpu_to_be32(vlan | MLX4_VLAN_VALID);

	err = mlx4_SET_PORT_vlan_table(dev, port, table->entries);
	if (unlikely(err)) {
		mlx4_warn(dev, "Failed adding vlan: %u\n", vlan);
		table->refs[free] = 0;
		table->entries[free] = 0;
		goto out;
	}

	*index = free;
	++table->total;
out:
	up(&table->vlan_sem);
	return err;
}
EXPORT_SYMBOL_GPL(mlx4_register_vlan);

void mlx4_unregister_vlan(struct mlx4_dev *dev, u8 port, int index)
{
	struct mlx4_vlan_table *table =
	&mlx4_priv(dev)->port[port - 1].vlan_table;

	down(&table->vlan_sem);
	if (!table->refs[index]) {
		mlx4_warn(dev, "No vlan entry for index %d\n", index);
		goto out;
	}
	if (--table->refs[index]) {
		mlx4_dbg(dev, "Have more references for index %d,"
			 "no need to modify vlan table\n", index);
		goto out;
	}
	table->entries[index] = 0;
	mlx4_SET_PORT_vlan_table(dev, port, table->entries);
	--table->total;
out:
	up(&table->vlan_sem);
}
EXPORT_SYMBOL_GPL(mlx4_unregister_vlan);

int mlx4_SET_PORT(struct mlx4_dev *dev, u8 port, 
	int reset_qkey_viols, u32 cap_mask)
{
	struct mlx4_cmd_mailbox *mailbox;
	int err;
	u8 is_eth = (dev->caps.port_type[port] == MLX4_PORT_TYPE_ETH) ? 1 : 0;

	mailbox = mlx4_alloc_cmd_mailbox(dev);
	if (IS_ERR(mailbox))
		return PTR_ERR(mailbox);

	memset(mailbox->buf, 0, 256);
	if (dev->flags & MLX4_FLAG_OLD_PORT_CMDS) {
		*(u8 *) mailbox->buf	     = (u8)(!!reset_qkey_viols << 6);
		((__be32 *) mailbox->buf)[2] = cpu_to_be32(cap_mask);
	} else {
		((u8 *) mailbox->buf)[3]     = (u8)!!reset_qkey_viols;
		((__be32 *) mailbox->buf)[1] = cpu_to_be32(cap_mask);
	}
	
	if (is_eth) {
		((u8 *) mailbox->buf)[3] = 7;
		((__be16 *) mailbox->buf)[3] =
			cpu_to_be16(dev->caps.eth_mtu_cap[port] +
				    ETH_HLEN + ETH_FCS_LEN);
		((__be16 *) mailbox->buf)[4] = cpu_to_be16(1 << 15);
		((__be16 *) mailbox->buf)[6] = cpu_to_be16(1 << 15);
	} else {
	/* IB PORT */
		
		((__be32 *) mailbox->buf)[0] |= cpu_to_be32((1 << 22) | (1 << 21) | (5 << 12) | (2 << 4));
		/* Set mmc bit no. 22 to allow MAX_MTU_SIZE change */
		/* Set mvc bit no. 21 to allow vl_cap change /*
		/* Always set MAX_MTU_SIZE == 4K for MLX4 - bits 12:15 */
		/* Change vl_cap to be maximum 4 - bits 5:7 */
	}
	err = mlx4_cmd(dev, mailbox->dma.da, port, is_eth, MLX4_CMD_SET_PORT,
		       MLX4_CMD_TIME_CLASS_B);

	mlx4_free_cmd_mailbox(dev, mailbox);
	return err;
}

